#include "swait.h"

struct SWAIT_DATA
{
    HANDLE    *pHandles ;
    DWORD    dwMilliseconds ;
    BOOL    bWaitAll ;
} ;

DWORD WINAPI SWaitThreadFunc( SWAIT_DATA *pData )
{
    DWORD dwIndex ;

    dwIndex = 0 ;

    // Find the number of handles to wait on
    while( pData->pHandles[ dwIndex ] && (dwIndex < 64) ) 
        dwIndex ++;

    // Wait on Handles
    dwIndex = WaitForMultipleObjects( dwIndex, pData->pHandles, 
        pData->bWaitAll, pData->dwMilliseconds ) ;

    // Delete paramater data struct (SWAIT_DATA)
    delete pData ;

    return     dwIndex ;
}

DWORD SuperWaitForMultipleObjects( DWORD nCount, 
    CONST HANDLE *lpHandles, BOOL bWaitAll, DWORD dwMilliseconds )
{    
    HANDLE        *phThreads ;
    HANDLE        *phWaitHandles ;
    HANDLE        hSemaphore ;
    int            nNumThreads, nLoop, nLoop2  ;
    DWORD        dwMainWait, nIndex, dwObject ;
    SWAIT_DATA    *pData ;
    BOOL        bFailed, bAbandoned, bTimeout ;

    // If we can use the standard function, then we will and 
    // translate return values
    if( nCount <= 64 )
    {
        dwObject = WaitForMultipleObjects( nCount, lpHandles, 
            bWaitAll, dwMilliseconds ) ;
        
        if( dwObject == WAIT_FAILED )
        {
            dwObject = SWAIT_FAILED ;
        }else if( dwObject == WAIT_TIMEOUT )
        {
            dwObject = SWAIT_TIMEOUT ;
        }else if( dwObject >= WAIT_ABANDONED_0 )
        {
            dwObject = dwObject + 
                (SWAIT_ABANDONED_0 - WAIT_ABANDONED_0) ;
        }

        return dwObject ;
    }

    // Get the number of threads required to do a super wait
    nNumThreads = nCount / 63 ;
    nNumThreads += (nCount%63?1:0) ;

    // At this point we are not supporting more than 64 
    // threads here... this allows up to 4032 wait objects
    if( nNumThreads > 64 ) 
        return SWAIT_FAILED ;

    // Create buffers for threads and wait handles
    phThreads = new HANDLE[ nNumThreads ] ;
    phWaitHandles = new HANDLE[ nNumThreads * 64 ] ;

    // Create the semaphore used to release "other" threads
    // If this is a "WaitAll" wait then this semaphore is not 
    // necessary, so we create it with a starting count of 
    // nNumThreads
    hSemaphore = CreateSemaphore( NULL, bWaitAll?nNumThreads:0, 
        nNumThreads, NULL ) ;
    nIndex = 0 ;

    // Create our threads... passing parameters via 
    // SWAIT_DATA structures
    for( nLoop = 0 ; nLoop < nNumThreads ; nLoop++ )
    {
        phWaitHandles[ nLoop*64 ] = hSemaphore ;
        for( nLoop2 = 1 ; nLoop2 < 64 ; nLoop2++ )
        {
            phWaitHandles[ nLoop*64+nLoop2 ] = NULL ;
            if( nIndex < nCount )
            {
                phWaitHandles[ nLoop*64+nLoop2 ] = 
                lpHandles[ nIndex++ ] ;
            }
        }
        pData = new SWAIT_DATA ;
        pData->pHandles = &(phWaitHandles[64*nLoop]) ;
        pData->dwMilliseconds = dwMilliseconds ;
        pData->bWaitAll = bWaitAll ;
        phThreads[ nLoop ] = CreateThread( NULL, 256, 
            (LPTHREAD_START_ROUTINE)SWaitThreadFunc, 
            pData, 0, &dwMainWait ); 
    }

    // Wait for one or all of the threads to end
    dwMainWait = WaitForMultipleObjects( nNumThreads, phThreads, 
        bWaitAll, INFINITE ) ;
    // Get thread(s)'s exit code, and build a return value from it
    if( bWaitAll )
    {        
        if( dwMainWait == WAIT_FAILED )
        {
            dwObject = SWAIT_FAILED ;
        }else 
        {
            bFailed = bAbandoned = bTimeout = FALSE ;
            for( nLoop = 0 ; nLoop < nNumThreads ; nLoop++ )
            {
                GetExitCodeThread( phThreads[nLoop], &dwObject ) ;
                bFailed = bFailed | (dwObject == WAIT_FAILED ) ;
                bAbandoned = bAbandoned | 
                    (dwObject == WAIT_ABANDONED_0 ) ;
                bTimeout = bTimeout | (dwObject == WAIT_TIMEOUT ) ;
            }
            if( bFailed )
            {
                dwObject = SWAIT_FAILED ;
            }else if( bTimeout )
            {
                dwObject = SWAIT_TIMEOUT ;
            }else if( bAbandoned ) 
            {
                dwObject = SWAIT_ABANDONED_0 ;
            }else
            {
                dwObject = SWAIT_OBJECT_0 ;
            }

        }
    }else
    {                
        if( dwMainWait != WAIT_FAILED && GetExitCodeThread( 
            phThreads[dwMainWait-WAIT_OBJECT_0], &dwObject ) )
        {
            if( dwObject == WAIT_FAILED )
            {
                dwObject = SWAIT_FAILED ;
            }else if( dwObject == WAIT_TIMEOUT )
            {
                dwObject = SWAIT_TIMEOUT ;
            }else if( dwObject >= WAIT_ABANDONED_0 )
            {                
                dwObject = (SWAIT_ABANDONED_0) + dwMainWait * 63 +
                    (dwObject-1) ;
            }else 
            {
                dwObject = (SWAIT_OBJECT_0) + dwMainWait * 63 +
                    (dwObject-1) ;
            }
        }else
        {
            dwObject = SWAIT_FAILED ;
        }
    }    

    //  Allow all threads to end their wait
    ReleaseSemaphore( hSemaphore, nNumThreads, NULL ) ;

    //  Wait for all the threads to terminate
    WaitForMultipleObjects( nNumThreads, phThreads, 
        TRUE, INFINITE ) ;

    //  Clean up
    CloseHandle( hSemaphore ) ;
    for( nLoop = 0 ; nLoop < nNumThreads ; nLoop++ )
    {
        CloseHandle( phThreads[ nLoop ] ) ;
    }

    delete phThreads ;    
    delete phWaitHandles ;    

    return dwObject ;
}
